import os

import yaml
from app.constants import config_path

def generate_default_config():
    #there is a design issue - no two models can have the same name, or this doesn't work
    #but im not fixing that, who cares, your fault for importing two different models with the same name lol

    return {
        "model_sources": {  # API hosters (gemini, ollama, openai etc)
            "gemini": {"api_key": None}
        },
        "models": [],  # Specific models (llama3, gemini flash 2.0, etc)
        "default_model": None  # The model to use by default
    }

class ConfigManager:
    """
    Handles the configuration file, safely reading and writing to the config file
    """

    def __init__(self):
        self.validate_config()

        # Load data after config is known to be OK
        try:
            with open(config_path, "r") as file:
                self.data = yaml.safe_load(file)

            if self.data is None:
                self.data = {}
        except FileNotFoundError:
            raise FileNotFoundError(f"Error: config file at path {config_path} not found.")

    def validate_config(self):
        """
        Ensures that the config file exists and is valid
        Automatically sets the config to the default config if invalidated
        """
        if not os.path.exists(config_path):  # Check config file exists
            self.reset_config()
            print(f"Config file not found, new config created at {config_path}")

    def reset_config(self, do_default_config = False) -> None:
        """
        Reset config to default values.
        Creates directory if it doesn't exist yet
        """
        if not os.path.exists(config_path):
            os.makedirs(os.path.dirname(config_path), exist_ok=True)

        with open(config_path, "w") as f:
            yaml.dump(generate_default_config(do_default_config), f, sort_keys=False)

    def get_config_variable(self, variable_name) -> any:
        """
        Safely get the value of a variable in the config
        :param variable_name: the variable of the config to access
        :return: the value of the variable, or None if not found
        """
        if variable_name in self.data:
            return self.data[variable_name]
        return None

    def set_config_variable(self, variable, new_value) -> None:
        """
        Writes a new value to the config file, overwriting an existing one or creating a new value
        :param variable: the name of the variable to set
        :param new_value: the value to set the variable to
        """
        self.data[variable] = new_value
        with open(config_path, 'w') as f:
            yaml.dump(self.data, f, default_flow_style=False, sort_keys=False, allow_unicode=True, Dumper=yaml.Dumper)


import yaml

from .message import Message
from .model import Model


class Chat:
    """
    Stores all messages sent by the user or LLM agents
    """
    def __init__(self):
        self._messages = []
        self.system_prompt = ""

    def add_message(self, message : Message):
        if len(message.content.strip()) != 0:
            self._messages.append(message.to_dict())

    def remove_last_message(self):
        self._messages.pop()

    def clear(self):
        self._messages = []

    def get_gemini_payload(self):
        """
        :return: Returns the messages formatted for gemini usage. Does not work for older models due to system instruction TODO
        """
        if len(self._messages) == 0:
            return None

        content = []
        system_message_parts = [{"text" : self.system_prompt}]
        for m in self._messages:
            if m["role"] != "system":
                role = "model" if m["role"] == "assistant" else m["role"]
                content.append({"role": role, "parts": [{"text": m["content"]}]})
            else:
                system_message_parts.append({"text" : m["content"]})

        if len(system_message_parts) != 0:
            return {"system_instruction" : {"parts" : system_message_parts}, "contents" : content}
        else:
            return {"contents" : content}

    def load(self, path: str, display_messages: bool = False, confirm_load=False) ->  None:
        """
        Loads in chat data from a given chat .yaml file
        :param path: the path to the chat .yaml file
        :param display_messages: whether to display a list of all messages after loading
        :param confirm_load: whether to print a confirmation message of loading or not
        """
        try:
            with open(path, "r") as file:
                data = yaml.safe_load(file)

            if data is None:
                return

            self._messages = data["messages"]
            self.system_prompt = data["system_prompt"]

            if confirm_load:
                print(f"Successfully loaded messages from {path}")

            if display_messages:
                self.display_chat_data()
        except FileNotFoundError:
            raise FileNotFoundError(f"Error: {path} not found.")
        except TypeError:
            raise TypeError(f"Error: {path} contains an invalid format.")

    def export(self, chat_source : str, model : Model, confirm_export : bool = False) -> None:
        """
        Save a Chat class into a .yaml file
        :param chat_source: the path to the .yaml file to save the chat into
        :param model: the model that is currently being used for this chat
        :param confirm_export: whether to print a confirmation message of exporting or not
        """
        data = {
                "model" : model.model_name,
                "system_prompt": self.system_prompt,
                "messages": self._messages
                }

        with open(chat_source, "w") as file:
            yaml.dump(data, file, default_flow_style=False, sort_keys=False)

        if confirm_export:
            print(f"Successfully exported messages to {chat_source}")

    def display_chat_data(self):
        chat = ""
        for i,m in enumerate(self._messages):
            chat += f"{m["role"]}: {m["content"]}\n"

        print(f"System prompt: {self.system_prompt}\n")
        print(chat)

    def set_system_prompt(self, system_prompt : str):
        self.system_prompt = system_prompt


import json
from typing import Iterator

import requests

from .util import connected_to_internet


class ModelError(Exception):
    """Base exception class for errors related to model output"""
    pass

class InvalidModelException(Exception):
    """Exception raised when an invalid model is requested"""
    pass

class NoInternetException(InvalidModelException):
    """Exception raised when an attempting to request a model with no internet connection"""
    pass

class InvalidAPIKeyException(InvalidModelException):
    """Exception raised when an invalid api key is given"""
    pass

class ModelParameters:
    """
    A class to store and manage parameters
    """
    def __init__(
        self,
        temperature: float,
        top_k: int,
        top_p: float,
        num_predict : int
    ):
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p
        self.num_predict = num_predict

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the parameters
        """
        return {}

class GeminiModelParameters(ModelParameters):
    """
    A class to store and manage parameters for Ollama model generation.
    """
    def __init__(
        self,
        temperature: float = 0.8,
        top_k: int = 40,
        top_p: float = 0.9,
        num_predict: int = 128
    ):
        """
        Initialize the parameters. Default is the ollama defaults
        :param temperature: Controls randomness. Higher is more creative, lower is more deterministic.  Default: 0.8.
        :param num_predict: Maximum number of tokens to generate.
                     Default: 128. Use -1 for infinite, -2 to fill context.
        :param top_k: Samples from the k most likely next tokens.   Default: 40.
        :param top_p: Samples from the smallest set of tokens whose cumulative probability exceeds p.   Default: 0.9.
        """
        super().__init__(temperature, top_k, top_p, num_predict)

    def to_dict(self) -> dict:
        """
        Returns the parameters as a JSON dictionary for generationConfig for gemini
        Specific to requests/curl api system
        """
        return {
            "temperature": self.temperature,
            "topP": self.top_p,
            "topK": self.top_k,
            "maxOutputTokens": self.num_predict
        }

class Model:
    """
    Generic model to host LLM
    """
    def __init__(self, model_name: str, api_key : str, debug: bool = False, parameters: ModelParameters = None):
        """
        :param model_name: The name of the model to use.
        :param api_key : The api key to use for this model
        :param debug: Display debug messages or not. Defaults to False
        :param parameters: The model parameters to use
        :raises InvalidModelException: Raised when an error occurs retrieving model
        """
        self.model_name = model_name
        self.api_key = api_key
        self.debug = debug
        self.parameters = parameters

    def raise_model_exists(self):
        """
        Checks if model name exists from API
        Also checks if no internet connection found
        :raises InvalidModelException: Raised when an invalid model name is given.
        :raises NoInternetException: Raised when no internet connection is found while attempting to retrieve model
        :raises InvalidAPIKeyException: Raised when an invalid API key is given.
        """
        pass

    def invoke(self, prompt : str = None) -> str:
        """
        Invoke the LLM with the given prompt
        :param prompt: The prompt to invoke
        :return: The output message
        :raises ModelError: when an error occurs
        """
        pass

class LocalModel(Model):
    """Model specifically for local models (mainly, and probably exclusively, ollama)"""
    def __init__(self, model_name: str, debug: bool = False, parameters: ModelParameters = None):
        super().__init__(model_name, None, debug, parameters)

class GeminiModel(Model):
    """
    Interface to send prompts to Gemini models with Google API. Uses REST API over python SDK for finer-grained control
    """
    def __init__(self, model_name: str, api_key : str, debug: bool = False, parameters: GeminiModelParameters = None):
        """
        :param model_name: The name of the model to use. Should be existing ollama model
        :param api_key: The API key for accessing the model
        :param debug: Display debug messages or not. Defaults to False
        :param parameters: The model parameters to use
        :raises InvalidModelException: Raised when an error occurs retrieving model
        """
        super().__init__(model_name, api_key, debug, parameters)
        self.raise_model_exists()

    def get_response(self, prompt : str = None, payload : dict = None, stream : bool = False, timeout : int = 60):
        """
        Gets the response from the model
        :param prompt: An optional parameter for the current message being sent
        :param payload: A dictionary representing the JSON payload to be sent.
                        Contains data like chat history
        :param stream: If True, the response body will not be downloaded immediately.
                       Will not handle streaming-related errors. Defaults to False.
        :param timeout: The timeout in seconds for the request. Defaults to 60.
        :return: The requests.Response object on a successful call.
        :raises ModelError: Raised for any network-level errors (e.g., connection, timeout) or for non-2xx HTTP status codes.
        """
        if stream: #Pick the URL for streaming or not streaming
            url = f"https://generativelanguage.googleapis.com/v1beta/models/{self.model_name}:streamGenerateContent?alt=sse&key={self.api_key}"
        else:
            url = f"https://generativelanguage.googleapis.com/v1beta/models/{self.model_name}:generateContent?key={self.api_key}"

        if payload is None:
            prompt = "" if prompt is None else prompt
            payload = {"contents": [{"parts": [{"text": prompt}]}]}

        # Add custom parameters if they exist
        if self.parameters is not None:
            payload["generationConfig"] = self.parameters.to_dict()

        headers = {"Content-Type": "application/json"}

        try:
            response = requests.post(url, headers=headers, data=json.dumps(payload), stream=stream, timeout=timeout)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise ModelError(f"HTTP error: {e.response.status_code} - {e.response.text}")
        except requests.exceptions.RequestException as e:
            raise ModelError(f"Error calling Gemini API: {e}")

        return response

    def raise_model_exists(self):
        """
        Checks if model name exists in gemini API
        :raises InvalidModelException: Raised when an invalid model name is given. Also raised when there is no internet.
        :raises NoInternetException: Raised when no internet connection is found while attempting to retrieve model
        :raises InvalidAPIKeyException: Raised when an invalid API key is given.
        """

        if not connected_to_internet():
            raise NoInternetException(f"Error fetching model : Not connected to the internet")

        try:
            models_url = f"https://generativelanguage.googleapis.com/v1beta/models?key={self.api_key}"
            response = requests.get(models_url)
            response.raise_for_status()

            models_data = response.json()

            available_model_ids = []
            if "models" in models_data:
                for model in models_data["models"]:
                    model_id = model.get("name", "").split("/")[-1]
                    if model_id:
                        available_model_ids.append(model_id)

            if self.model_name not in available_model_ids:
                raise InvalidModelException(f"Model {self.model_name} is not a valid model name.\n"
                                            f"Valid model names are: {", ".join(available_model_ids)}")

        except requests.exceptions.RequestException as e:
            raise InvalidAPIKeyException(f"Error fetching model: {e}\n"
                                        f"Check your API key is correct")
        except Exception as e:
            raise InvalidModelException(f"An unexpected error occurred while fetching models: {e}")

    def invoke(self, prompt : str = None, payload : dict = None) -> str:
        """
        Invoke the LLM with the given prompt
        :param prompt: The prompt to invoke
        :param payload: The payload containing any extra data (e.g. chat history)
        :return: The output message
        :raises ModelError: when an error occurs
        """
        response = self.get_response(prompt = prompt, payload = payload, stream = False)
        try:
            return response.json()['candidates'][0]['content']['parts'][0]['text']
        except (KeyError, IndexError) as e:
             raise ModelError(f"Error formatting the json response {e}")

    def stream(self, prompt : str = None, payload : dict = None) -> Iterator[str]:
        """
        Streamed the LLM output with the given prompt
        :param prompt: The prompt to invoke, optional
        :param payload: The payload containing any extra data (e.g. chat history)
        :return: The output message
        :raises ModelError: when an error occurs
        """
        response = self.get_response(prompt = prompt, payload = payload, stream = True)
        try:
            for chunk in response.iter_lines():
                if chunk:
                    decoded_chunk = chunk.decode("utf-8")
                    if decoded_chunk.startswith('data: '):
                        parsed_chunk = decoded_chunk[len('data: '):]
                        json_chunk = json.loads(parsed_chunk)
                        yield json_chunk['candidates'][0]['content']['parts'][0].get('text', '')
        except Exception as e:
            raise ModelError(f"An unexpected error occurred: {e}")

    def invoke_chat(self, chat_payload : dict[str, dict[str, list]]):
        """
        Get a single LLM output from a given Chat history
        :param chat_payload: the chat data formatted for usage with the gemini API (chat.get_gemini_payload())
        :return: a stream of the output message
        TODO Error messsage when invalid payload for debug, also include in stream
        TODO better yet, refactor this silly code
        """
        return self.invoke(payload = chat_payload)

    def stream_chat(self, chat_payload : dict[str, dict[str, list]]):
        """
        Get a streamed LLM output from a given Chat history
        :param chat_payload: the chat data formatted for usage with the gemini API (chat.get_gemini_payload())
        :return: Text stream for the LLM
        """
        return self.stream(payload = chat_payload)


import logging

from .template import Template

valid_roles = ["user", "system", "assistant"]

class Message:
    def __init__(self, role : str, content : str):
        self.role = role.lower().strip()
        self.content = content.strip()

        if self.role not in valid_roles:
            logging.warning(f"Role {self.role} is not a valid role. Ensure roles are one of {valid_roles}")

    def to_dict(self):
        return {"role" : self.role, "content" : self.content}

class FormattedMessage(Message):
    def __init__(self, role : str, content : str, format_data : dict):
        Message.__init__(self, role, Template(content, missing_behaviour = "warn").format(format_data))


from typing import Optional

import typer

from app import chat_core
from app.constants import *
from app.util import pretty_terminal_table

from app.config_manager import ConfigManager
from app.model_manager import ModelManager
from app.chat_manager import ChatManager


class App:
    def __init__(self):
        self.config_manager = ConfigManager()
        self.model_manager = ModelManager(self.config_manager)
        self.chat_manager = ChatManager()

        self.app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]},
                          no_args_is_help=True,
                          add_completion=False)

        self.config_app = typer.Typer(no_args_is_help=True, help = "Manage configuration settings")
        self.system_prompt_app = typer.Typer(no_args_is_help=True, help="Manage system prompts for chats ")
        self.model_app = typer.Typer(no_args_is_help=True, help="Manage the settings for the current model ")

        self.app.add_typer(self.config_app, name="config")
        self.app.add_typer(self.system_prompt_app, name = "systemprompt")
        self.app.add_typer(self.model_app, name="model")

        self._register_commands()

    def _register_commands(self):
        # main
        self.app.command(name="start")(self.start)
        self.app.command(name="once")(self.once)
        self.app.command(name="list")(self.list_chats)
        self.app.command(name="delete")(self.delete_chat)

        # config
        self.config_app.command(name="find")(self.config_find_command)
        self.config_app.command(name="reset")(self.config_reset_command)

        # system prompt
        self.system_prompt_app.command(name="set")(self.set_system_prompt)
        self.system_prompt_app.command(name="show")(self.show_system_prompt)
        self.system_prompt_app.command(name="load")(self.load_system_prompt)

        # model settings
        self.model_app.command(name="select")(self.select_model)
        self.model_app.command(name="add")(self.add_model)
        self.model_app.command(name="remove")(self.remove_model)
        self.model_app.command(name="setapi")(self.set_api_key)
        self.model_app.command(name="list")(self.list_models)

    # -------------- main commands -------------- #

    def start(self,
              chat_name: Optional[str] = typer.Argument(None, help="Optional name of the chat history to start."),
              no_stream: bool = typer.Option(False, "--nostream", is_flag=True, help = "Disable streaming"),
              no_markdown: bool = typer.Option(False, "--nomarkdown", is_flag=True, help = "Disable markdown printing")):
        """
        Starts the chat, optionally giving the name of the chat history to start.
        """
        model = self.model_manager.get_default_model(self.config_manager)
        if model is not None:
            chat_path = self.chat_manager.select_chat(chat_name)
            chat_core.start_chat(chat_path, model, not no_stream, not no_markdown)

    def once(self,
             message: Optional[list[str]] = typer.Argument(None, help = "The message to send to the LLM"),
             chat_name: Optional[str] = typer.Option(None, "--chat", help="Specify the chat history name to export"),
             no_stream: bool = typer.Option(False, "--nostream", is_flag=True, help="Disable streaming"),
             no_markdown: bool = typer.Option(False, "--nomarkdown", is_flag=True, help="Disable markdown printing")):
        """
        Send a single chat message.
        """
        model = self.model_manager.get_default_model(self.config_manager)
        if model is None:
            return

        if message is None:
            full_message = input("Enter a chat message >> ")
        else:
            full_message = " ".join(message)

        chat_source = None
        if chat_name is not None:
            chat_source = self.chat_manager.select_chat(chat_name)

        chat_core.single_message(full_message, model, chat_source, not no_stream, not no_markdown)

    def list_chats(self):
        """
        List all existing chats.
        """
        self.chat_manager.list_chats()

    def delete_chat(self, chat_name: str = typer.Argument(help="The name of the chat to delete"),):
        """
        Deletes a given chat
        """
        self.chat_manager.delete_chat(chat_name)

    # -------------- system prompt commands -------------- #

    def set_system_prompt(self,
                          system_prompt: list[str] = typer.Argument(help="The system prompt to set"),
                          chat_name: Optional[str] = typer.Option(None, "--chat", help="Specify the chat history name to load the system prompt.")):
        """
        Sets the system prompt for the given chat history.
        """
        if chat_name is None:
            print(f"Error: The --chat option is required to specify which chat history to modify.")
            raise typer.Exit(code=1)

        # Join the list of strings into a single system prompt string
        full_system_prompt = " ".join(system_prompt)
        self.chat_manager.set_system_prompt(chat_name, full_system_prompt)

    def show_system_prompt(self, chat_name: str = typer.Argument(help="Specify the chat history name to display the system prompt of.")):
        """
        Displays the system prompt for a given chat
        """
        chat_data = self.chat_manager.get_chat_data(chat_name)
        if chat_data is not None:
            print(f"System prompt for chat '{chat_name}':\n\"{chat_data["system_prompt"].strip()}\"")

    def load_system_prompt(self,
                          path: str = typer.Argument(help="Filepath for the system prompt to load"),
                          chat_name: Optional[str] = typer.Option(None, "--chat", help="Specify the chat history name to load the system prompt.")):
        """
        Sets the system prompt for the given chat history.
        """
        if chat_name is None:
            print(f"Error: The --chat option is required to specify which chat history to modify.")
            raise typer.Exit(code=1)

        try:
            with open(path, 'r', encoding='utf-8') as f:
                system_prompt = f.read()

        except FileNotFoundError:
            print(f"Error: The file '{path}' was not found.")
        except Exception as e:
            print(f"An error occurred reading the data: {e}")

        self.chat_manager.set_system_prompt(chat_name, system_prompt)

    # -------------- config commands -------------- #

    def config_find_command(self):
        """
        Displays the location of the config file and the chat data
        """
        print(f"Config is in {config_path}")
        print(f"Chat data is in {data_path}")

    def config_reset_command(self, preset_config: bool = typer.Option(False, "--default", "-d", is_flag=True,
                             help = "Use an alternative config that attempts to find a .env file in the __main__.py location for api keys. Really only exists for debugging, feel free to ignore.")):
        """
        Resets the config file to default
        """
        self.config_manager.reset_config(preset_config)
        print(f"Config successfully reset to default in {config_path}")

    # -------------- model commands -------------- #

    def select_model(self, model_name: str = typer.Argument(help="The name of the model to use")):
        """
        Sets the given model to be the default model
        """
        if self.model_manager.is_model_in_config(model_name):
            self.config_manager.set_config_variable("default_model", model_name)
            print(f"Set default model to {model_name}")
        else:
            print(f"{model_name} is not a valid model name")
            print(f"Check models with {cli_keyword} model list, "
                  f"or add a model with {cli_keyword} model add <model_name> <model_source>")


    def add_model(self,
                  model_name: str = typer.Argument(help="The name of the model to use"),
                  model_source: str = typer.Argument(help=f"The source of the model. Type {cli_keyword} model to list all model sources")):
        """
        Add a model to the collection of models
        """
        if self.model_manager.validate_model(model_name, model_source):
            models = self.config_manager.get_config_variable("models")
            model = {"name" : model_name, "source" : model_source}
            if model not in models:
                models.append(model)
                self.config_manager.set_config_variable("models", models)
                print(f"Successfully created new model {model_name}")
            else:
                print(f"Model {model_name} from {model_source} already exists")

    def remove_model(self, model_name: str = typer.Argument(help="The name of the model to remove")):
        """
        Removes a given model from the selection.
        """
        if self.model_manager.is_model_in_config(model_name):
            default = self.config_manager.get_config_variable("default_model")
            if default == model_name:
                print(f"Failed to delete. {model_name} is the selected model.")
                print(f"Select a different model with {cli_keyword} model select <model_name> before deleting model {model_name}")
            else:
                models = self.config_manager.get_config_variable("models")
                new_models = []
                for model in models:
                    if model["name"] != model_name:
                        new_models.append(model)
                        break
                self.config_manager.set_config_variable("models", new_models)
                print(f"Removed model {model_name}")
        else:
            print(f"{model_name} is not a valid model name")
            print(f"Check models with {cli_keyword} model list, "
                  f"or add a model with {cli_keyword} model add <model_name> <model_source>")

    def set_api_key(self,
                  model_source: str = typer.Argument(help=f"The source to add the API for. Type {cli_keyword} model to list all model sources"),
                  api_key: str = typer.Argument(help="The api key to set for this source")):
        """
        Set the API key for a model source
        """
        sources = self.config_manager.get_config_variable("model_sources")
        if model_source not in sources:
            print(f"The source {model_source} is not in the list of supported model sources")
            print(f"Type '{cli_keyword} model' to find a list of supported model sources")
            return

        sources[model_source]["api_key"] = api_key
        self.config_manager.set_config_variable("model_sources", sources)

        print(f"Successfully set new api key for source {model_source}")

    def list_models(self):
        """
        List all current created model data, and model sources
        """
        models = self.config_manager.get_config_variable("models")
        model_sources = self.config_manager.get_config_variable("model_sources")
        if len(model_sources) > 0:
            print("\nSupported model sources:")
            for source in model_sources:
                print(f" - {source}")
            print()
        else:
            print(f"No model sources. (Reset your config with {cli_keyword} config reset)\n")

        if len(models) > 0:
            default_model = self.config_manager.get_config_variable("default_model")

            rows = [[model["name"], model["source"]] for model in models]
            print(f"Default model : {default_model}\n")
            pretty_terminal_table(rows, ["Model name", "Model source"])
            print()
        else:
            print("No models found.")

def run():
    app = App()
    app.app()

if __name__ == "__main__":
    run()


import datetime
import os
from pathlib import Path

import yaml

from app.constants import data_path, cli_keyword
from app.util import pretty_terminal_table


class ChatManager:
    """
    Manages finding, loading and saving chat data
    """
    def __init__(self):
        self.validate_chats()

    def validate_chats(self) -> None:
        """
        Ensures that the stored data exists and the chats are valid
        Creates a new data directory if it doesn't already exist
        """
        if not os.path.exists(data_path):
            os.makedirs(data_path, exist_ok=True)
            print(f"Chat data directory not found, new directory created at {data_path}")

    def get_chat_paths(self):
        """
        :return: a list of all chat .yaml files in the data directory
        """
        return list(Path(data_path).glob("*.yaml"))

    def get_chat_path(self, chat_name: str):
        """
        Get the chat path for a specific chat
        :param chat_name: the name of the chat get the filepath for
        :return: The path for the chat if it exists, None otherwise
        """
        chat_path = os.path.join(data_path, chat_name + ".yaml")
        if not os.path.exists(chat_path):
            return None
        return chat_path

    def get_most_recent_chat(self):
        """
        :returns The most recent chat path
        :throws ValueError if no chat exists (max([]))
        """
        return max(self.get_chat_paths(), key=lambda f: f.stat().st_mtime)

    def select_chat(self, chat_name : str = None):
        """
        Given a chat name, attempts to find the chat with the given name and returns the path
        If chat name is not an existing chat, creates a new chat with the given name
        If no chat name given, selects the most recently accessed chat
        If no chats found, automatically creates a new empty chat
        :param chat_name: the name of the chat to access
        :return: The path of the chat to access
        """
        if chat_name is None: #No name given
            if len(self.get_chat_paths()) == 0:
                print("No existing chats found, creating a new chat 'chat' ")
                chat_path = os.path.join(data_path, "chat.yaml")

            else:
                print("Selecting most recent chat")
                return self.get_most_recent_chat()
        else: #Name has been given
            chat_path = os.path.join(data_path, chat_name + ".yaml")

        Path(chat_path).touch(exist_ok=True)  # Create an empty yaml chat file

        return chat_path

    def get_chat_data(self, chat_name):
        """
        Gets the yaml data as a dict from a chat
        :param chat_name: the chat name to get the data for
        :return: a dictionary containing the deserialized yaml file, or None if no chat found
        """
        chat_path = self.get_chat_path(chat_name)

        if chat_path is None:
            print(f"No chat with name {chat_name} found. "
                  f"\nCreate the chat first with chat start {chat_name}, or type chat list to list available chats")
            return None

        with open(chat_path, "r") as file:
            data = yaml.safe_load(file)

        return data

    def set_system_prompt(self, chat_name : str, system_prompt : str):
        """
        Sets the system prompt for a chat
        If the chat cannot be found, informs the user
        :param chat_name: the name of the chat to set the prompt for
        :param system_prompt: the system prompt
        """
        chat_path = self.get_chat_path(chat_name)
        data = self.get_chat_data(chat_name)
        if data is None:
            return

        data["system_prompt"] = system_prompt

        with open(chat_path, "w") as file:
            yaml.dump(data, file, default_flow_style=False, sort_keys=False)

        print(f"Successfully updated chat '{chat_name}' with new system prompt ")

    def delete_chat(self, chat_name : str):
        """
        Deletes a chat after user confirmation
        :param chat_name: the name of the chat to delete
        """
        chat_path = self.get_chat_path(chat_name)
        if chat_path is None:
            print(f"No chat with name {chat_name} found.")
            return

        option = input(f"Are you sure you want to delete the chat {chat_name}? "f"\ny/n >> ")
        if option.strip() == "y":
            os.remove(chat_path)
            print(f"Successfully removed chat in {chat_path}")


    def list_chats(self):
        """
        Display all chats and some metadata about them
        """
        chat_paths = self.get_chat_paths()
        if len(chat_paths) == 0:
            print(f"No chats found. Start a new chat with '{cli_keyword} start <chat_name>'")
            return

        #setup table for pretty print
        column_names = ["Chat name", "Last used", "Model"]
        rows = []
        for chat_path in chat_paths:
            with open(chat_path, "r") as file:
                data = yaml.safe_load(file)

            chat_name = chat_path.name.removesuffix(".yaml")
            last_used = chat_path.stat().st_mtime
            last_used_datetime = datetime.datetime.fromtimestamp(last_used).strftime("%d/%m/%Y %H:%M")
            model = data["model"]

            rows.append([chat_name, last_used_datetime, model])

        pretty_terminal_table(rows, column_names, padding = 5)


import os

from ai_core.model import GeminiModel  # , OllamaModel
from platformdirs import user_config_dir, user_data_dir

cli_keyword = "chat" #The command alias the program uses
program_name = "ai_chat" #The directory name for the program

config_path = os.path.join(user_config_dir(program_name), "config.yaml")
data_path = user_data_dir(program_name)

# TODO add more sources
MODEL_SOURCES = {
    "gemini" : GeminiModel,
    #"ollama": OllamaModel}
}



import re

class Template:
    def __init__(self, prompt : str,
                prefix : str = "{{",
                suffix : str = "}}",
                missing_behaviour : str = "empty",
                default_arguments : dict = None):
        """
        Generic template class
        :param prompt: The prompt containing tokens to be replaced.
        :param prefix: The string that marks the beginning of a token (default: "{{").
        :param suffix: The string that marks the end of a token (default: "}}").
        :param missing_behaviour: The behaviour to replace missing token
                                  - "empty" replaces missing tokens with empty strings. (default)
                                  - "ignore" does not do anything to the missing tokens.
                                  - "warn" does do anything, but sends a warning message. TODO logging
        :param default_arguments: The default token/values to be replaced. Defaults to None.
        """
        self.prompt = prompt
        self.prefix = prefix
        self.suffix = suffix
        self.missing_behaviour = missing_behaviour
        self.default_arguments = default_arguments

    def add_default(self, token, value) -> None:
        """
        Adds a default token to the template.
        :param token: the token to be added.
        :param value: the value of the default token to replace it
        """
        if self.default_arguments is None:
            self.default_arguments = {}
        self.default_arguments[token] = value

    def format(self, _arguments : dict = None, **kwargs) -> str:
        """
        Formats the template with a given set of tokens.
        Example use: template.format({"token" : "replacement", ...}) and/or template.format(token = "replacement", ...)
        :param _arguments: The values to replace the tokens with as a dictionary. Optional
        :param kwargs: Arbitrary keyword args where each keyword is a token name and its value is the replacement.
        :return: The formatted string.
        """

        arguments = {}
        if _arguments:
            arguments.update(_arguments)
        arguments.update(kwargs)

        tokens_found = re.findall(f"{re.escape(self.prefix)}(.*?){re.escape(self.suffix)}", self.prompt)
        if not tokens_found:  # No tokens to replace
            return self.prompt

        formatted_prompt = self.prompt
        for token in set(tokens_found):
            full_token = f"{self.prefix}{token}{self.suffix}"
            if token in arguments:
                formatted_prompt = formatted_prompt.replace(full_token, str(arguments[token]))
            elif self.default_arguments and token in self.default_arguments:
                formatted_prompt = formatted_prompt.replace(full_token, str(self.default_arguments[token]))
            else:
                match self.missing_behaviour:
                    case "empty": formatted_prompt = formatted_prompt.replace(full_token, "")
                    case "warn" : print(f"WARNING: Token {token} was not be replaced")
                    case "ignore" : pass

        return formatted_prompt

    def partial_format(self, _arguments: dict = None, **kwargs) -> "Template":
        """
        Formats the template with a given set of tokens, but returns a new template with this prompt instead of a string
        :param _arguments: The values to replace the tokens with as a dictionary. Optional
        :param kwargs: Arbitrary keyword args where each keyword is a token name and its value is the replacement.
        :return: The partially formatted template
        """
        formatted_string = format(_arguments, **kwargs)
        return Template(formatted_string, self.prefix, self.suffix, self.missing_behaviour, self.default_arguments)

    def expected_tokens(self) -> set[str]:
        """
        Gets all tokens present in the template.
        :return: A set of all tokens that may be replaced in the template
        """
        return set(re.findall(f"{re.escape(self.prefix)}(.*?){re.escape(self.suffix)}", self.prompt))


from ai_core.chat import Chat
from ai_core.message import Message
from ai_core.model import Model
from ai_core.util import output_stream, markdown_print

help_message = """
type quit/bye/exit to quit (saves your chat)
type clear to clear the chat history (asks for confirmation)
type retry to regenerate the previous AI response
type save to save the chat so far
type systemprompt/system to display and/or change the system prompt
type help to display this message
"""

def output_response(chat : Chat, model : Model, do_stream : bool, do_markdown : bool):
    if do_stream:
        stream = model.stream_chat(chat.get_gemini_payload())
        response = output_stream(stream, do_markdown=do_markdown)
    else:
        response = model.invoke_chat(chat.get_gemini_payload())
        if do_markdown:
            markdown_print(response.strip())
        else:
            print(response.strip())

    return response

def single_message(message : str, model : Model, chat_source = None, do_stream = True, do_markdown = True):
    chat = Chat()

    if chat_source is not None: #load cha
        chat.load(chat_source, False)

    chat.add_message(Message("user", message))

    response = output_response(chat, model, do_stream, do_markdown)

    if chat_source is not None:
        chat.add_message(Message("assistant", response))
        chat.export(chat_source, model, False)

def start_chat(chat_source : str, model : Model, do_stream = True, do_markdown = True):
    chat = Chat()

    chat.load(chat_source, False)
    while True:
        prompt = input(">> ")

        if len(prompt) != 0:
            match = True
            match prompt.lower().strip():
                case "h" | "help":
                    print(help_message)
                case "quit" | "q" | "bye" | "exit":
                    chat.export(chat_source, model)
                    break
                case "clear":
                    if input("Are you sure you want to clear the chat history? (y/n) > ").lower() == "y":
                        chat.clear()
                        print("Chat history cleared")
                case "retry":
                    print("Regenerating a new response...")
                    chat.remove_last_message()
                    match = False
                case "save":
                    chat.export(chat_source, model, confirm_export = True)
                case "system" | "systemprompt":
                    print(f"Current system prompt:\n\"{chat.system_prompt.strip()}\"\n")
                    system_prompt = input("Input the new system prompt for this chat, type nothing to cancel"
                                          "\nprompt >> ").strip()
                    if system_prompt != "":
                        chat.set_system_prompt(system_prompt)
                        print("Successfully set system prompt")
                case _:
                    match = False

            if match:
                continue

            chat.add_message(Message("user", prompt))

        response = output_response(chat, model, do_stream, do_markdown)

        chat.add_message(Message("assistant", response))


def pretty_terminal_table(rows : list[list], column_names : list,  padding : int = 3, border : str = None, row_prefix = ""):
    column_widths = []
    columns = list(map(list, zip(*rows)))

    for i, column in enumerate(columns):
        width = len(column_names[i])
        for value in column:
            width = max(width, len(str(value)))
        column_widths.append(width)

    table = [column_names]
    for row in rows:
        table.append(row)

    for row in table:
        current_column = 0
        for item in row:
            value = str(item)
            if current_column == 0:
                value = row_prefix + value

            total_padding = column_widths[current_column] - len(str(item))
            if border is None or current_column == len(columns) - 1: #Dont draw border after final column
                total_padding += padding
                print(value + " " * total_padding, end = "")
            else:
                total_padding += padding//2
                print(value + " " * total_padding + border + " " * (padding//2) , end = "")

            current_column += 1
        print()


from typing import Type

from ai_core.model import Model, LocalModel, InvalidModelException, InvalidAPIKeyException

from app.constants import MODEL_SOURCES, cli_keyword


class ModelManager:
    def __init__(self, config_manager):
        self.model_source_data = config_manager.get_config_variable("model_sources")
        self.saved_models = config_manager.get_config_variable("models")
        self.default_model_name = config_manager.get_config_variable("default_model")

        if self.model_source_data is None or len(self.model_source_data) == 0:
            print("Warning - No model sources found in config. Ensure config file is valid.")
            return


        # Validate all existing saved models
        models = config_manager.get_config_variable("models")
        for model in models:
            result = self.validate_model(model["name"], model["source"], display_errors = False)
            if not result:
                print(f"Warning - an error occurred validating a saved model. ({model})\n"
                      f"Check the config files are up to date and valid, and ensure API keys are valid")
                break

    def validate_model(self, model_name : str, model_source : str, display_errors = True) -> bool:
        """
        Validates a model from a source to see if it exists or not
        :param model_name: the name of the model
        :param model_source: the source of the model
        :param display_errors: whether to display error messages or not
        :return: True if the model exists, False otherwise
        """
        model_source = model_source.strip() #Validate model source
        if model_source not in self.model_source_data:
            if display_errors:
                print(f"{model_source} is not a currently supported model source. Supported sources:")
                for source in self.model_source_data:
                    print(f" - {source}")
            return False

        #Attempt to create model
        model = self.get_model(model_name, model_source, display_errors = display_errors)

        return model is not None

    def is_model_in_config(self, model_name):
        return any(model["name"] == model_name for model in self.saved_models)

    def get_model(self, model_name : str, model_source : str, display_errors = True) -> None | Model | LocalModel:
        """
        Returns a Model class, given name and model source.
        Used for checking if a given model from a source actually can exist
        Used in model add <name> <source>
        :param model_name: the name of the model.
        :param model_source: the api source of the model (e.g. gemini, ollama, etc)
        :param display_errors: whether to display error messages or not
        :return: The Model class corresponding to the saved model name/source, or None on error
        """

        model: Type[Model] = MODEL_SOURCES[model_source]

        try:
            if issubclass(model, LocalModel):
                return model(model_name)
            else:
                api_key = self.model_source_data[model_source]["api_key"]
                return model(model_name, api_key)
        except InvalidAPIKeyException:
            if display_errors:
                print(f"The API key for source {model_source} is invalid.")
                print(f"Set a valid API key for {model_source} with {cli_keyword} model setapi <model_source> <api_key>")
            return None
        except InvalidModelException as e:
            if display_errors:
                print(f"An error occurred validating this model: {e}")
            return None

    def get_model_from_config(self, model_name : str) -> Model | None:
        """
        Returns a Model class, given only the name, from the saved model data
        :param model_name: the name of the model.
        :return: The Model class corresponding to the saved model name/source.
        """
        model_source = None
        for saved_model in self.saved_models:
            if saved_model["name"] == model_name:
                model_source = saved_model["source"]

        if model_source not in MODEL_SOURCES:
            print(f"Model {model_name} has an invalid source ({model_source}). Ensure config is valid")
            return None

        return self.get_model(model_name, model_source)

    def select_new_default_model(self, config_manager):
        """
        Fallback method to select the first model from the saved models
        :return: the name of the model to select, or None if no models can be selected
        """
        if len(self.saved_models) == 0:
            return None
        else:
            self.default_model_name = self.saved_models[0]["name"]
            config_manager.set_config_variable("default_model", self.default_model_name)
            print(f"Auto-selected new default model to {self.default_model_name}")
            return self.default_model_name

    def get_default_model(self, config_manager) -> Model | None:
        if self.default_model_name is None:
            self.default_model_name = self.select_new_default_model(config_manager)

        if self.default_model_name is None: #If  still none after selecting new default, this means no models exist
            print(f"Error! No models exist yet! Create a new model with {cli_keyword} model add <model_name> <model_source>")
            return None

        return self.get_model_from_config(self.default_model_name)




import urllib.request

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown


def markdown_print(text, end="\n", do_markdown=True):
    console = Console()

    if do_markdown:
        if "\n" not in text.strip():
            console.print(text, markup=True, end=end)
        else:
            console.print(Markdown(text))
            if end != "\n":
                console.print(end, end="")
    else:
        print(text, end=end)

def connected_to_internet() -> bool:
    try:
        urllib.request.urlopen('https://google.com')
        return True
    except urllib.error.URLError:
        return False

def output_stream(stream, do_markdown=True):
    """
    Reads content from a string stream token by token, updating a live
    """
    console = Console()
    capture = ""

    if do_markdown:
        with Live(console=console, refresh_per_second=64) as live:
            for raw_chunk in stream:
                capture += raw_chunk
                markdown_renderable = Markdown(capture)
                live.update(markdown_renderable)
    else:
        for raw_chunk in stream:
            capture += raw_chunk
            print(raw_chunk, end="")

    if not capture.endswith("\n"):
        print("")

    return capture.strip()
